Unit IPX;

(*****************************************)
(*    Written by Alexei A. Frounze (c)   *)
(*        on 27 of September 1999        *)
(*                                       *)
(*     Target:  Real/Protected mode      *)
(*  Compiler: Borland/Turbo Pascal 7.0   *)
(*                                       *)
(* E-mail: alexfru@chat.ru               *)
(* Homepage: http://www.chat.ru/~alexfru *)
(*****************************************)

Interface

Const
  MaxUsers = 10;                                { Maximum players }

Type
  PFullAddress  = ^TFullAddress;
  TFullAddress  = Record
    NetAddress  : Array[0..3] of Byte;          { LAN address (reversed) }
    Node        : Array[0..5] of Byte;          { Node address (reversed), i.e. this computer number }
    Socket      : Word                          { Socket number (reversed) }
  End;
  PIPXHeader = ^TIPXHeader;
  TIPXHeader = Record
    Checksum,                                   { ? }
    Length        : Word;                       { Size of this header + data size }
    TransportCtrl,                              { ? }
    PacketType    : Byte;                       { 4 = IPX }
    Destination,                                { Receiver address }
    Source        : TFullAddress                { Sender address }
  End;
  PECB = ^TECB;
  TECB = Record
    LinkAddr,                                   { ? }
    ESRAddr         : Pointer;                  { ? }
    InUse,                                      { name says :) }
    CompletionCode  : Byte;                     { name says :) }
    Socket          : Word;                     { Socket number (reveersed) }
    IPXWorkSpace    : Array[0..3] of Byte;      { ? }
    DriverWorkSpace : Array[0..11] of Byte;     { ? }
    ImmediateAddr   : Array[0..5] of Byte;      { Receiver/Sender }
    FragmentCount   : Word;                     { 2 for our program }
    Fragment1Addr   : Pointer;                  { IPX header address }
    Fragment1Size   : Word;                     { IPX header size }
    Fragment2Addr   : Pointer;                  { Data address }
    Fragment2Size   : Word                      { Data size }
  End;
  PIOBuf =^TIOBuf;
  TIOBuf = Record
    IPXHeader : TIPXHeader;
    ECB       : TECB;
    Data      : Array [0..511] of Byte          { Fixed size !!! }
  End;

Var
  _Addr : TFullAddress;                         { This computer's net addr }

Function AllocRMBufs : Boolean;
{ Called automatically on startup }

Function FreeRMBufs : Boolean;
{ You should it call before "End." }

Function Is_IPX_Installed : Boolean;
{ Guess what it does :) }

Procedure IPX_Get_Full_Address (Var Addr : TFullAddress);
{ Returns your full network address }

Function IPX_Open_Socket (CloseByCall : Boolean; Var Socket : Word) : Byte;
{ Opens the "Socket" socket }

Procedure IPX_Close_Socket (Socket : Word);
{ Closes the "Socket" socket }

Procedure IPX_Relinquish_Control;
{ You have to call it once per each game cycle in order to make }
{ IPX driver working correctly }

Procedure IPX_Send_Data (Var Data; Var ToNode);
{ Sends 512-bytes data block to the "ToNode" node (6 bytes) }
{ In order to do a broadcast sending, "ToNode" should be filled }
{ with $FFFFFFFFFFFF }

Procedure IPX_Prepare_For_Receiving (UserFirst, UserLast : Word);
{ Prepears receiving of (UserLast-UserFirst+1) data blocks }
{ Note: UserFirst <= UserLast < MaxUsers }

Function IPX_Receive_Data (UserNo : Word; Var Data; Var FromNode) : Boolean;
{ Returns True, if receives a 512-bytes data block }
{ Note: UserNo should be within UserFirst...UserLast range (see above) }
{ Sender node is returned in "FromNode" var (6 bytes) }

Implementation

{$IFDEF DPMI}

Uses DOS, WinAPI;

Type
  PRegs = ^TRegs;
  TRegs = Record
     EDI, ESI, EBP, Res, EBX, EDX, ECX, EAX : LongInt;
     Flags, ES, DS, FS, GS, IP, CS, SP, SS : Word;
  End;

Var
  PRMFullAddress : PFullAddress;
  PRMECB         : PECB;
  RMFullAddressHandle,
  RMFullAddressSeg,
  RMECBHandle,
  RMECBSeg       : LongInt;
  RMIOBufHandle,
  RMIOBufSeg     : LongInt;

{$ELSE}

Uses DOS;

{$ENDIF}

Var
  PRMIOBuf : PIOBuf;                            { I/O Buffers }

(*************** DPMI Support functions ***************)

Function AllocRMBufs : Boolean;
Begin
{$IFDEF DPMI}
  RMFullAddressHandle := GlobalDOSAlloc (SizeOf(TFullAddress));
  RMFullAddressSeg := RMFullAddressHandle shr 16;
  PRMFullAddress := Ptr (RMFullAddressHandle and $FFFF, 0);

  RMECBHandle := GlobalDOSAlloc (SizeOf(TECB));
  RMECBSeg := RMECBHandle shr 16;
  PRMECB := Ptr (RMECBHandle and $FFFF, 0);

  RMIOBufHandle := GlobalDOSAlloc (SizeOf(TIOBuf)*(MaxUsers+1));
  RMIOBufSeg := RMIOBufHandle shr 16;
  PRMIOBuf := Ptr (RMIOBufHandle and $FFFF, 0);
{$ELSE}
  GetMem (PRMIOBuf, SizeOf(TIOBuf)*(MaxUsers+1));
{$ENDIF}
  AllocRMBufs := True
End;

Function FreeRMBufs : Boolean;
Begin
{$IFDEF DPMI}
  GlobalDOSFree (RMFullAddressHandle);
  GlobalDOSFree (RMECBHandle);
  GlobalDOSFree (RMIOBufHandle);
{$ELSE}
  FreeMem (PRMIOBuf, SizeOf(TIOBuf)*(MaxUsers+1));
{$ENDIF}
  FreeRMBufs := True
End;

{$IFDEF DPMI}
Function RealIntr (Int : Byte; Var RMR : Registers) : Boolean;
Var
  R : Registers;
  Regs : TRegs;
Begin
  With RMR, Regs do Begin
    EAX := AX;
    EBX := BX;
    ECX := CX;
    EDX := DX;
    EBP := BP;
    ESI := SI;
    EDI := DI;
    Regs.Flags := RMR.Flags;
    Regs.DS := RMR.DS;
    Regs.ES := RMR.ES;
    SP := 0;
    SS := 0
  End;
  R.AX := $300;
  R.BL := Int;
  R.BH := 0;
  R.CX := 0;
  R.ES := Seg (Regs);
  R.DI := Ofs (Regs);
  Intr ($31, R);
  If R.Flags and FCarry <> 0 then RealIntr := False;
  With RMR, Regs do Begin
    AX := EAX;
    BX := EBX;
    CX := ECX;
    DX := EDX;
    BP := EBP;
    SI := ESI;
    DI := EDI;
    RMR.Flags := Regs.Flags;
    RMR.DS := Regs.DS;
    RMR.ES := Regs.ES
  End;
  RealIntr := True
End;
{$ENDIF}

(*************** DPMI Support functions ***************)

                           (* IPX ROUTINES *)

Function Is_IPX_Installed : Boolean; Assembler;
Asm
  Mov   AX, 7A00h
  Int   2Fh
  And   AL, 1
End;

Function IPX_Max_Packet_Size : Word;
Var R : Registers;
Begin
{$IFDEF DPMI}
  R.BX := $1A;
  RealIntr ($7A, R);
  IPX_Max_Packet_Size := R.AX
{$ELSE}
  R.BX := $1A;
  Intr ($7A, R);
  IPX_Max_Packet_Size := R.AX
{$ENDIF}
End;

Procedure IPX_Get_Full_Address (Var Addr : TFullAddress);
Var R : Registers;
Begin
{$IFDEF DPMI}
  R.BX := 9;
  R.ES := RMFullAddressSeg;
  R.SI := 0;
  RealIntr ($7A, R);
  Move (PRMFullAddress^, Addr, SizeOf(TFullAddress));
{$ELSE}
  R.BX := 9;
  R.ES := Seg(Addr);
  R.SI := Ofs(Addr);
  Intr ($7A, R);
{$ENDIF}
End;

Function IPX_Open_Socket (CloseByCall : Boolean; Var Socket : Word) : Byte;
Var R : Registers;
Begin
{$IFDEF DPMI}
  R.BX := 0;
  R.AL := Byte (CloseByCall);
  R.DX := ((Socket and $FF) shl 8) or (Socket shr 8);
  RealIntr ($7A, R);
  Socket := ((R.DX and $FF) shl 8) or (R.DX shr 8);
  IPX_Open_Socket := R.AL
{$ELSE}
  R.BX := 0;
  R.AL := Byte (CloseByCall);
  R.DX := ((Socket and $FF) shl 8) or (Socket shr 8);
  Intr ($7A, R);
  Socket := ((R.DX and $FF) shl 8) or (R.DX shr 8);
  IPX_Open_Socket := R.AL
{$ENDIF}
End;

Procedure IPX_Close_Socket (Socket : Word);
Var R : Registers;
Begin
{$IFDEF DPMI}
  R.BX := 1;
  R.DX := ((Socket and $FF) shl 8) or (Socket shr 8);
  RealIntr ($7A, R);
{$ELSE}
  R.BX := 1;
  R.DX := ((Socket and $FF) shl 8) or (Socket shr 8);
  Intr ($7A, R);
{$ENDIF}
End;

Procedure IPX_Send_Packet (Var ECBS : TECB);
Var R : Registers;
Begin
{$IFDEF DPMI}
  Move (ECBS, PRMECB^, SizeOf(TECB));
  R.BX := 3;
  R.ES := RMECBSeg;
  R.SI := 0;
  RealIntr ($7A, R);
{$ELSE}
  R.BX := 3;
  R.ES := Seg(ECBS);
  R.SI := Ofs(ECBS);
  Intr ($7A, R);
{$ENDIF}
End;

Function IPX_Listen_For_Packet (Var ECBR : TECB) : Byte;
Var R : Registers;
Begin
{$IFDEF DPMI}
  R.BX := 4;
  R.ES := RMIOBufSeg;
  R.SI := Ofs(ECBR);
  RealIntr ($7A, R);
  IPX_Listen_For_Packet := R.AL
{$ELSE}
  R.BX := 4;
  R.ES := Seg(ECBR);
  R.SI := Ofs(ECBR);
  Intr ($7A, R);
  IPX_Listen_For_Packet := R.AL
{$ENDIF}
End;

Procedure IPX_Relinquish_Control;
Var R : Registers;
Begin
{$IFDEF DPMI}
  R.BX := $A;
  RealIntr ($7A, R);
{$ELSE}
  R.BX := $A;
  Intr ($7A, R);
{$ENDIF}
End;

Procedure IPX_Send_Data (Var Data; Var ToNode);
Var P : PIOBuf;
Begin
  P := PRMIOBuf;
  Inc (Word(P), MaxUsers*SizeOf(TIOBuf));
  with P^.IPXHeader do Begin
    Checksum := 0;                                      { ? }
    Length := SizeOf(TIPXHeader)+512;                   { Length of packet }
    TransportCtrl := 0;                                 { ? }
    PacketType := 4;                                    { IPX }
    Destination.NetAddress := _Addr.NetAddress;         { Our LAN }
    Move (ToNode, Destination.Node, SizeOf(_Addr.Node));{ Broadcast/Certain recipient }
    Destination.Socket := _Addr.Socket;                 { Current socket }
    Move (_Addr, Source, SizeOf(TFullAddress))          { Sender - this computer }
  End;
  with P^.ECB do Begin
    LinkAddr := Nil;                                    { ? }
    ESRAddr := Nil;                                     { ? }
    Socket := _Addr.Socket;                             { Current socket }
    FillChar (IPXWorkSpace, 4, 0);                      { ? }
    FillChar (DriverWorkSpace, 12, 0);                  { ? }
    Move (ToNode, ImmediateAddr, SizeOf(_Addr.Node));   { Broadcast/Certain recipient }
    FragmentCount := 2;                                 { 2 information blocks }
{$IFDEF DPMI}
    Fragment1Addr := Ptr (RMIOBufSeg, Word(P));
{$ELSE}
    Fragment1Addr := @P^.IPXHeader;                     { 1st block is }
{$ENDIF}
    Fragment1Size := SizeOf(TIPXHeader);                { IPX header }
{$IFDEF DPMI}
    Fragment2Addr := Ptr (RMIOBufSeg, Word(P)+SizeOf(TIPXHeader)+SizeOf(TECB));
{$ELSE}
    Fragment2Addr := @P^.Data;                          { 2nd block is }
{$ENDIF}
    Fragment2Size := 512;                               { data block we send }
  End;
  Move (Data, P^.Data, 512);
  IPX_Send_Packet (P^.ECB)                              { Send packet }
End;

Procedure IPX_Prepare_For_Receiving (UserFirst, UserLast : Word);
Var
  P : PIOBuf;
  I : Word;
Begin
  If (UserFirst > UserLast) or (UserLast >= MaxUsers) then Exit;
  P := PRMIOBuf;
  Inc (Word(P), UserFirst*SizeOf(TIOBuf));
  For I := UserFirst to UserLast do Begin
    with P^.ECB do Begin
      LinkAddr := Nil;                                    { ? }
      ESRAddr := Nil;                                     { ? }
      Socket := _Addr.Socket;                             { Current socket }
      FillChar (IPXWorkSpace, 4, 0);                      { ? }
      FillChar (DriverWorkSpace, 12, 0);                  { ? }
      FillChar (ImmediateAddr, 6, $FF);                   { Broadcast }
      FragmentCount := 2;                                 { 2 information blocks }
{$IFDEF DPMI}
      Fragment1Addr := Ptr (RMIOBufSeg, Word(P));
{$ELSE}
      Fragment1Addr := @P^.IPXHeader;                     { 1st block is }
{$ENDIF}
      Fragment1Size := SizeOf(TIPXHeader);                { IPX header }
{$IFDEF DPMI}
      Fragment2Addr := Ptr (RMIOBufSeg, Word(P)+SizeOf(TIPXHeader)+SizeOf(TECB));
{$ELSE}
      Fragment2Addr := @P^.Data;                          { 2nd block is }
{$ENDIF}
      Fragment2Size := 512;                               { data block we send }
    End;
    IPX_Listen_For_Packet (P^.ECB);                       { Listen for packet }
    Inc (Word(P), SizeOf(TIOBuf))
  End
End;

Function IPX_Receive_Data (UserNo : Word; Var Data; Var FromNode) : Boolean;
Var
  P : PIOBuf;
  I : Word;
Begin
  P := PRMIOBuf;
  Inc (Word(P), UserNo*SizeOf(TIOBuf));
  If P^.ECB.InUse = 0 then Begin
    Move (P^.Data, Data, 512);
    Move (P^.ECB.ImmediateAddr, FromNode, SizeOf(_Addr.Node));
    IPX_Prepare_For_Receiving (UserNo, UserNo);
    IPX_Receive_Data := True
  End
  Else
    IPX_Receive_Data := False
End;

Begin
  AllocRMBufs;
End.
